class MalformedRegionException(Exception):
    pass


class SPIUnavailableException(Exception):
    pass


class ECCMalfunctionException(Exception):
    pass


class GatewayMFRFileNotFoundException(Exception):
    pass


class MinerFailedToFetchMacAddress(Exception):
    pass


class MinerFailedToFetchEthernetAddress(Exception):
    pass


class UnknownVariantException(Exception):
    pass


class UnknownVariantAttributeException(Exception):
    pass


class GatewayMFRExecutionException(Exception):
    pass


class GatewayMFRInvalidVersion(Exception):
    pass


class UnsupportedGatewayMfrVersion(Exception):
    pass



import base58
import grpc

from hm_pyhelper.protos import blockchain_txn_add_gateway_v1_pb2, \
    local_pb2_grpc, local_pb2, region_pb2, gateway_staking_mode_pb2
from hm_pyhelper.gateway_grpc.exceptions import MinerMalformedAddGatewayTxn

from hm_pyhelper.logger import get_logger

LOGGER = get_logger(__name__)


def decode_pub_key(encoded_key: bytes) -> str:
    # Addresses returned by the RPC response are missing a leading
    # byte for the version. The version is currently always 0.
    # https://github.com/helium/helium-js/blob/8d5cb76e156fb80de6fc80f239b43e3872c7b7d7/packages/crypto/src/Address.ts#L64
    version_byte = b'\x00'

    # Convert binary address to base58
    complete_key = version_byte + encoded_key
    decoded_key = base58.b58encode_check(complete_key).decode()
    return decoded_key


class GatewayClient(object):
    '''
    GatewayClient wraps grpc api provided by helium gateway-rs
    It provides some convenience methods to support the old api
    to limit breaking changes.
    Direct interaction with the grpc api can be achieved by
    using GatewayClient.stub.<api>

    All methods might return grpc pass through exceptions.
    '''

    def __init__(self, url='helium-miner:4467'):
        self._url = url
        self._channel = grpc.insecure_channel(url)
        self._channel.subscribe(self._connect_state_handler)
        self.stub = local_pb2_grpc.apiStub(self._channel)

    def _connect_state_handler(self, state):
        if state == grpc.ChannelConnectivity.SHUTDOWN:
            LOGGER.error('GRPC Channel shutdown : irrecoverable error')

    def __enter__(self):
        return self

    def __exit__(self, _, _2, _3):
        self._channel.close()

    def api_stub(self):
        return self.stub

    def get_region_enum(self) -> int:
        '''
        Returns the current configured region of the gateway.
        If not asserted or set in settings, defaults to 0 (US915)
        ref: https://github.com/helium/proto/blob/master/src/region.proto
        '''
        return self.stub.region(local_pb2.region_req()).region

    def get_region(self) -> str:
        '''
        Returns the current configured region of the gateway.
        If not asserted or set in settings, defaults to 0 (US915)
        '''
        region_id = self.get_region_enum()
        return region_pb2.region.Name(region_id)

    def get_pubkey(self) -> str:
        '''
        Returns decoded public key of the gateway
        '''
        encoded_key = self.stub.pubkey(local_pb2.pubkey_req()).address
        return decode_pub_key(encoded_key)

    def get_summary(self) -> dict:
        '''
        Returns a dict with following information
        {
            "region": str
                configured region eg. "US915",
            "key": str
                gateway/device public key
        }
        '''
        return {
            'region': self.get_region(),
            'key': self.get_pubkey()
        }

    def create_add_gateway_txn(self, owner_address: str, payer_address: str,
                               staking_mode: gateway_staking_mode_pb2.gateway_staking_mode
                               = gateway_staking_mode_pb2.gateway_staking_mode.light
                               ) -> bytes:
        """
        Invokes the txn_add_gateway RPC endpoint on the gateway and returns
        the same payload that the smartphone app traditionally expects.
        https://docs.helium.com/mine-hnt/full-hotspots/become-a-maker/hotspot-integration-testing/#generate-an-add-hotspot-transaction

        Parameters:
            - owner_address: The address of the account that owns the gateway.
            - payer_address: The address of the account that will pay for the
                             transaction. This will typically be the
                             maker/Nebra's account.
            - staking_mode: The staking mode of the gateway.
                            ref:
                            https://github.com/helium/proto/blob/master/src/service/local.proto#L38
        """
        # base58 gives version number as first byte. Get rid of it.
        owner = base58.b58decode_check(owner_address)[1:]
        payer = base58.b58decode_check(payer_address)[1:]
        response = self.stub.add_gateway(local_pb2.add_gateway_req(
            owner=owner,
            payer=payer,
            staking_mode=staking_mode
        ))
        return response.add_gateway_txn


def get_address_from_add_gateway_txn(add_gateway_txn:
                                     blockchain_txn_add_gateway_v1_pb2,
                                     address_type: str,
                                     expected_address: str = None):
    """
    Deserializes specified field in the blockchain_txn_add_gateway_v1_pb2
    protobuf to a base58 Helium address.

    Params:
        - add_gateway_txn: The blockchain_txn_add_gateway_v1_pb2 to
                           inspect.
        - address_type: 'owner', 'gateway', or 'payer'.
        - expected_address (optional): Value we expect to be returned.

    Raises:
        MinerMalformedAddGatewayTxn if expected_address supplied and
        does not match the return value.
    """

    # Addresses returned by the RPC response are missing a leading
    # byte for the version. The version is currently always 0.
    # https://github.com/helium/helium-js/blob/8d5cb76e156fb80de6fc80f239b43e3872c7b7d7/packages/crypto/src/Address.ts#L64
    version_byte = b'\x00'

    # Convert binary address to base58
    address_bytes = version_byte + getattr(add_gateway_txn, address_type)
    address = str(base58.b58encode_check(address_bytes), 'utf-8')

    # Ensure resulting address matches expectation
    is_expected_address_defined = expected_address is not None
    if is_expected_address_defined and address != expected_address:
        msg = f"Expected {address_type} address to be {expected_address}," + \
              f"but is {address}"
        raise MinerMalformedAddGatewayTxn(msg)

    return address



from unittest import TestCase
from hm_pyhelper.logger import get_logger, _log_format
import re
import logging


class TestLogger(TestCase):
    def test_get_logger(self):
        logger = get_logger(__name__)

        with self.assertLogs() as captured:
            logger.info("Hello world.")

        # check that there is only one log message
        self.assertEqual(len(captured.records), 1)
        record = captured.records[0]
        formatter = logging.Formatter(_log_format)
        formatted_output = formatter.format(record)

        # Do not check timestamp and filepath because those change
        # based on the environment and run time
        expected_partial_output_regex = re.escape(
            " - [INFO] - hm_pyhelper.tests.test_logger -" +
            " (test_logger.py).test_get_logger -- ")
        expected_output_regex = ".*" + \
            expected_partial_output_regex + ".*" + \
            " - Hello world."
        are_logs_correct = re.search(expected_output_regex, formatted_output)
        self.assertTrue(are_logs_correct)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: region.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0cregion.proto\x12\x06helium*\xf1\x02\n\x06region\x12\t\n\x05US915\x10\x00\x12\t\n\x05\x45U868\x10\x01\x12\t\n\x05\x45U433\x10\x02\x12\t\n\x05\x43N470\x10\x03\x12\r\n\x05\x43N779\x10\x04\x1a\x02\x08\x01\x12\t\n\x05\x41U915\x10\x05\x12\x0b\n\x07\x41S923_1\x10\x06\x12\t\n\x05KR920\x10\x07\x12\t\n\x05IN865\x10\x08\x12\x0b\n\x07\x41S923_2\x10\t\x12\x0b\n\x07\x41S923_3\x10\n\x12\x0b\n\x07\x41S923_4\x10\x0b\x12\x0c\n\x08\x41S923_1B\x10\x0c\x12\x0c\n\x08\x43\x44\x39\x30\x30_1A\x10\r\x12\t\n\x05RU864\x10\x0e\x12\x0b\n\x07\x45U868_A\x10\x0f\x12\x0b\n\x07\x45U868_B\x10\x10\x12\x0b\n\x07\x45U868_C\x10\x11\x12\x0b\n\x07\x45U868_D\x10\x12\x12\x0b\n\x07\x45U868_E\x10\x13\x12\x0b\n\x07\x45U868_F\x10\x14\x12\r\n\tAU915_SB1\x10\x15\x12\r\n\tAU915_SB2\x10\x16\x12\x0c\n\x08\x41S923_1A\x10\x17\x12\x0c\n\x08\x41S923_1C\x10\x18\x12\x0c\n\x08\x41S923_1D\x10\x19\x12\x0c\n\x08\x41S923_1E\x10\x1a\x12\x0c\n\x08\x41S923_1F\x10\x1b\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'region_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _REGION.values_by_name["CN779"]._options = None
  _REGION.values_by_name["CN779"]._serialized_options = b'\010\001'
  _REGION._serialized_start=25
  _REGION._serialized_end=394
# @@protoc_insertion_point(module_scope)



import unittest
import threading
import pytest
import mock
from time import sleep
from hm_pyhelper.lock_singleton import LockSingleton, ResourceBusyError, \
    lock_ecc


# https://gist.github.com/sbrugman/59b3535ebcd5aa0e2598293cfa58b6ab
@pytest.fixture(autouse=True, scope="function")
def error_on_raise_in_thread():
    """
    Replaces Thread with a a wrapper to record any exceptions and
    re-raise them after test execution. In case multiple threads
    raise exceptions only one will be raised.
    """
    last_exception = None

    class ThreadWrapper(threading.Thread):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        def run(self):
            try:
                super().run()
            except BaseException as e:
                nonlocal last_exception
                last_exception = e

    with mock.patch('threading.Thread', ThreadWrapper):
        yield
        if last_exception:
            raise last_exception


class TestLockSingleton(unittest.TestCase):
    def test_lock_singleton_simple(self):
        lock = LockSingleton()

        self.assertFalse(lock.locked())
        lock.acquire()

        # Do some work
        sleep(0.001)

        lock.release()
        self.assertFalse(lock.locked())

    def test_lock_singleton_timeout(self):
        """
        Start a slow running thread and then try to acquire a shared
        lock. Expect the acquire call to throw an exception.
        """

        def slow_task():
            sleep(0.1)
            lock.release()
            return True

        lock = LockSingleton()

        # ECC is going to be occupied by `slow_task`.
        slow_thread = threading.Thread(target=slow_task, daemon=True)
        lock.acquire()
        slow_thread.start()

        self.assertTrue(lock.locked())

        # Try to acquire the ECC that is occupied by `slow_task` currently.
        expected_exception = False
        try:
            lock.acquire(timeout=0.00001)
        except ResourceBusyError:
            expected_exception = True

        self.assertTrue(expected_exception)

    def test_lock_singleton_racing(self):
        def slow_task():
            lock = LockSingleton()

            lock.acquire()
            self.assertTrue(lock.locked())

            # Do some slow work
            print("Starting the slow task...")
            sleep(0.01)
            print("Finished the slow task!")

            lock.release()
            self.assertFalse(lock.locked())

        def fast_task():
            lock = LockSingleton()

            lock.acquire()
            self.assertTrue(lock.locked())

            # Do some slow work
            print("Starting the fast task...")
            sleep(0.001)
            print("Finished the fast task!")

            lock.release()
            self.assertFalse(lock.locked())

        slow_thread = threading.Thread(target=slow_task, daemon=True)
        fast_thread = threading.Thread(target=fast_task, daemon=True)

        # Start work
        print("\n")
        print("Initiated the slow task.")
        slow_thread.start()

        print("Initiated the fast task.")
        fast_thread.start()

        # Wait to finish
        try:
            fast_thread.join()
            slow_thread.join()
        except Exception as ex:
            print(ex)

    def test_lock_ecc_simple(self):
        @lock_ecc()
        def some_task():
            sleep(0.00001)

        some_task_thread = threading.Thread(target=some_task, daemon=True)
        some_task_thread.start()

        some_task_thread.join()

    def test_lock_ecc_timeout(self):
        @lock_ecc()
        def slow_task():
            sleep(0.01)

        @lock_ecc(timeout=0.001, raise_resource_busy_exception=True)
        def lock_with_timeout():
            pass

        slow_thread = threading.Thread(target=slow_task, daemon=True)
        slow_thread.start()

        expected_exception = False
        try:
            lock_with_timeout()
        except ResourceBusyError:
            expected_exception = True

        self.assertTrue(expected_exception)

    def test_lock_ecc_racing(self):
        @lock_ecc()
        def slow_task():
            # Do some slow work
            print("Starting the slow task...")
            sleep(0.003)
            print("Finished the slow task!")

        @lock_ecc()
        def fast_task():
            # Do some slow work
            print("Starting the fast task...")
            sleep(0.001)
            print("Finished the fast task!")

        slow_thread = threading.Thread(target=slow_task, daemon=True)
        fast_thread = threading.Thread(target=fast_task, daemon=True)

        # Start work
        print("\n")
        print("Initiated the slow task.")
        slow_thread.start()
        print("Initiated the fast task.")
        fast_thread.start()

        # Wait to finish
        fast_thread.join()
        slow_thread.join()

    def test_lock_ecc_forward_exception(self):
        @lock_ecc()
        def faulty_task():
            print("Starting the faulty task...")
            sleep(0.001)
            raise Exception("Intended faulty occurred!")

        expected_exception = False
        try:
            faulty_task()
        except Exception:
            expected_exception = True

        self.assertTrue(expected_exception)



"""
This module provides a set of functions to extract information about
the Single Board Computer in use.
It considers Balena environment variables as primary source of truth.
It also uses the device tree to extract information about the SBC.
"""

import os
from enum import Enum, auto
from collections import namedtuple

SBCInfo = namedtuple('SBCInfo', ['vendor_id', 'vendor_name', 'model_name'])


class DeviceVendorID(Enum):
    """
    Enum for device vendors.
    """
    INVALID = auto()
    ROCK_PI = auto()
    RASPBERRY_PI = auto()


# Pulled from
# https://www.balena.io/docs/reference/base-images/devicetypes/
BALENA_ENV_RASPBERRY_PI_MODELS = [
    'raspberry-pi',
    'raspberry-pi2',
    'raspberrypi3',
    'raspberrypi3-64',
    'raspberrypi4-64',
    'nebra-hnt',
    'raspberrypicm4-ioboard',
    'raspberrypi0-2w-64'
]

BALENA_ENV_ROCKPI_MODELS = ['rockpi-4b-rk3399']

BALENA_MODELS = {
    DeviceVendorID.ROCK_PI: BALENA_ENV_ROCKPI_MODELS,
    DeviceVendorID.RASPBERRY_PI: BALENA_ENV_RASPBERRY_PI_MODELS
}


def device_model():
    with open('/proc/device-tree/model', 'r') as f:
        return f.readline().strip()


def sbc_info() -> SBCInfo:
    '''
    return SBCInfo formed by reading '/proc/device-tree/model'
    '''
    sbc_info = SBCInfo(vendor_id=DeviceVendorID.INVALID, vendor_name='', model_name='')
    dev_model = device_model()
    if dev_model.lower().find('raspberry') >= 0:
        sbc_info = SBCInfo(vendor_id=DeviceVendorID.RASPBERRY_PI,
                           vendor_name='Raspberry Pi',
                           model_name=dev_model)
    elif dev_model.lower().find('rock') >= 0:
        sbc_info = SBCInfo(vendor_id=DeviceVendorID.ROCK_PI,
                           vendor_name='Radxa Rock Pi',
                           model_name=dev_model)
    return sbc_info


def is_sbc_type(device_id: DeviceVendorID) -> bool:
    '''
    Return true if the sbc matches the type supplied.
    '''
    device_type = os.getenv('BALENA_DEVICE_TYPE')

    # use device tree supplied model name if evn not set
    if not device_type:
        return sbc_info().vendor_id == device_id

    # honor env override
    return device_type in BALENA_MODELS.get(device_id, [])



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

import hm_pyhelper.protos.local_pb2 as local__pb2


class apiStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.pubkey = channel.unary_unary(
                '/helium.local.api/pubkey',
                request_serializer=local__pb2.pubkey_req.SerializeToString,
                response_deserializer=local__pb2.pubkey_res.FromString,
                )
        self.region = channel.unary_unary(
                '/helium.local.api/region',
                request_serializer=local__pb2.region_req.SerializeToString,
                response_deserializer=local__pb2.region_res.FromString,
                )
        self.router = channel.unary_unary(
                '/helium.local.api/router',
                request_serializer=local__pb2.router_req.SerializeToString,
                response_deserializer=local__pb2.router_res.FromString,
                )
        self.add_gateway = channel.unary_unary(
                '/helium.local.api/add_gateway',
                request_serializer=local__pb2.add_gateway_req.SerializeToString,
                response_deserializer=local__pb2.add_gateway_res.FromString,
                )


class apiServicer(object):
    """Missing associated documentation comment in .proto file."""

    def pubkey(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def region(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def router(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def add_gateway(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_apiServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'pubkey': grpc.unary_unary_rpc_method_handler(
                    servicer.pubkey,
                    request_deserializer=local__pb2.pubkey_req.FromString,
                    response_serializer=local__pb2.pubkey_res.SerializeToString,
            ),
            'region': grpc.unary_unary_rpc_method_handler(
                    servicer.region,
                    request_deserializer=local__pb2.region_req.FromString,
                    response_serializer=local__pb2.region_res.SerializeToString,
            ),
            'router': grpc.unary_unary_rpc_method_handler(
                    servicer.router,
                    request_deserializer=local__pb2.router_req.FromString,
                    response_serializer=local__pb2.router_res.SerializeToString,
            ),
            'add_gateway': grpc.unary_unary_rpc_method_handler(
                    servicer.add_gateway,
                    request_deserializer=local__pb2.add_gateway_req.FromString,
                    response_serializer=local__pb2.add_gateway_res.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'helium.local.api', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class api(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def pubkey(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/helium.local.api/pubkey',
            local__pb2.pubkey_req.SerializeToString,
            local__pb2.pubkey_res.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def region(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/helium.local.api/region',
            local__pb2.region_req.SerializeToString,
            local__pb2.region_res.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def router(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/helium.local.api/router',
            local__pb2.router_req.SerializeToString,
            local__pb2.router_res.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def add_gateway(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/helium.local.api/add_gateway',
            local__pb2.add_gateway_req.SerializeToString,
            local__pb2.add_gateway_res.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: local.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


import hm_pyhelper.protos.gateway_staking_mode_pb2 as gateway__staking__mode__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0blocal.proto\x12\x0chelium.local\x1a\x1agateway_staking_mode.proto\"9\n\npubkey_res\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\x0c\x12\x1a\n\x12onboarding_address\x18\x02 \x01(\x0c\"\x0c\n\npubkey_req\")\n\tkeyed_uri\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\x0c\x12\x0b\n\x03uri\x18\x02 \x01(\t\"\x0c\n\nregion_req\"\x1c\n\nregion_res\x12\x0e\n\x06region\x18\x01 \x01(\x05\"\x0c\n\nrouter_req\",\n\nrouter_res\x12\x0b\n\x03uri\x18\x01 \x01(\t\x12\x11\n\tconnected\x18\x02 \x01(\x08\"c\n\x0f\x61\x64\x64_gateway_req\x12\r\n\x05owner\x18\x01 \x01(\x0c\x12\r\n\x05payer\x18\x02 \x01(\x0c\x12\x32\n\x0cstaking_mode\x18\x03 \x01(\x0e\x32\x1c.helium.gateway_staking_mode\"*\n\x0f\x61\x64\x64_gateway_res\x12\x17\n\x0f\x61\x64\x64_gateway_txn\x18\x01 \x01(\x0c\x32\x8c\x02\n\x03\x61pi\x12<\n\x06pubkey\x12\x18.helium.local.pubkey_req\x1a\x18.helium.local.pubkey_res\x12<\n\x06region\x12\x18.helium.local.region_req\x1a\x18.helium.local.region_res\x12<\n\x06router\x12\x18.helium.local.router_req\x1a\x18.helium.local.router_res\x12K\n\x0b\x61\x64\x64_gateway\x12\x1d.helium.local.add_gateway_req\x1a\x1d.helium.local.add_gateway_resb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'local_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _PUBKEY_RES._serialized_start=57
  _PUBKEY_RES._serialized_end=114
  _PUBKEY_REQ._serialized_start=116
  _PUBKEY_REQ._serialized_end=128
  _KEYED_URI._serialized_start=130
  _KEYED_URI._serialized_end=171
  _REGION_REQ._serialized_start=173
  _REGION_REQ._serialized_end=185
  _REGION_RES._serialized_start=187
  _REGION_RES._serialized_end=215
  _ROUTER_REQ._serialized_start=217
  _ROUTER_REQ._serialized_end=229
  _ROUTER_RES._serialized_start=231
  _ROUTER_RES._serialized_end=275
  _ADD_GATEWAY_REQ._serialized_start=277
  _ADD_GATEWAY_REQ._serialized_end=376
  _ADD_GATEWAY_RES._serialized_start=378
  _ADD_GATEWAY_RES._serialized_end=420
  _API._serialized_start=423
  _API._serialized_end=691
# @@protoc_insertion_point(module_scope)
